@@ -1,4 +1,4 @@
-from .core import Wtp
+from .core import Wtp, MAGIC_FIRST, MAGIC_LAST
 from .parser import NodeKind, WikiNode
 from .languages import ALL_LANGUAGES
 
@@ -7,4 +7,6 @@ __all__ = (
     "NodeKind",
     "WikiNode",
     "ALL_LANGUAGES",
+    "MAGIC_FIRST",  # Some applications with to use the same ranges
+    "MAGIC_LAST",
 )
