@@ -7,7 +7,9 @@ import os
 import re
 import sys
 import html
+import time
 import pickle
+import datetime
 import tempfile
 import traceback
 import collections
@@ -43,20 +45,38 @@ def phase2_page_handler(dt):
     ctx = _global_ctx
     autoload = _global_page_autoload
     model, title = dt
-    ctx.start_page(title)
-    if autoload:
-        data = ctx.read_by_title(title)
-        assert isinstance(data, str)
-    else:
-        data = None
-    try:
-        ret = _global_page_handler(model, title, data)
-        return True, ret
-    except Exception as e:
-        lst = traceback.format_exception(etype=type(e), value=e,
-                                         tb=e.__traceback__)
+    start_t = time.time()
 
-        return False, "=== EXCEPTION:\n" + "".join(lst)
+    # XXX Enable this to debug why extraction hangs.  This writes the path
+    # of each file being processed into /tmp/wiktextract-*.  Once a hang
+    # has been observed, these files contain page(s) that hang.  They should
+    # be checked before aborting the process, as an interrupt might delete
+    # them.
+    debug_hangs = False
+    try:
+        debug_path = "/tmp/wiktextract-{}".format(os.getpid())
+        with open(debug_path, "w") as f:
+            f.write(title + "\n")
+
+        ctx.start_page(title)
+        if autoload:
+            data = ctx.read_by_title(title)
+            assert isinstance(data, str)
+        else:
+            data = None
+        try:
+            ret = _global_page_handler(model, title, data)
+            return True, title, start_t, ret
+        except Exception as e:
+            lst = traceback.format_exception(etype=type(e), value=e,
+                                             tb=e.__traceback__)
+            msg = ("=== EXCEPTION while parsing page \"{}\":\n".format(title) +
+                   "".join(lst))
+            return False, title, start_t, msg
+
+    finally:
+        if debug_hangs:
+            os.remove(debug_path)
 
 
 class Wtp(object):
@@ -76,6 +96,9 @@ class Wtp(object):
         "errors",	 # List of error messages (cleared for each new page)
         "fullpage",	 # The unprocessed text of the current page (or None)
         "lua",		 # Lua runtime or None if not yet initialized
+        "lua_depth",     # Recursion depth in Lua calls
+        "lua_invoke",	 # Lua function used to invoke a Lua module
+        "lua_reset_env", # Function to reset Lua environment
         "lua_path",	 # Path to Lua modules
         "modules",	 # Lua code for defined Lua modules
         "need_pre_expand",  # Set of template names to be expanded before parse
@@ -95,6 +118,7 @@ class Wtp(object):
         "warnings",	 # List of warning messages (cleared for each new page)
         # Data for parsing
         "beginning_of_line", # Parser at beginning of line
+        "wsp_beginning_of_line",  # Parser at beginning of line + whitespace
         "linenum",	 # Current line number
         "pre_parse",	 # XXX is pre-parsing still needed?
         "parser_stack",	 # Parser stack
@@ -124,6 +148,9 @@ class Wtp(object):
         self.section = None
         self.subsection = None
         self.lua = None
+        self.lua_invoke = None
+        self.lua_reset_env = None
+        self.lua_depth = 0
         self.quiet = quiet
         self.rev_ht = {}
         self.expand_stack = []
@@ -168,7 +195,8 @@ class Wtp(object):
         self.need_pre_expand = None
         self.cache_file_old = False
         # Add predefined templates
-        self.templates["!"] = "&vert;"
+        self.templates["!"] = "|"
+        self.templates["!-"] = "|-"
         self.templates[self._canonicalize_template_name("((")] = \
             "&lbrace;&lbrace;"  # {{((}}
         self.templates[self._canonicalize_template_name("))")] = \
@@ -208,8 +236,8 @@ class Wtp(object):
                                  NodeKind.LEVEL6):
                     if not node.args:
                         continue
-                    lst = filter(lambda x: x if isinstance(x, str) else "???",
-                                 node.args[0])
+                    lst = map(lambda x: x if isinstance(x, str) else "???",
+                              node.args[0])
                     title = "".join(lst)
                     titles.append(title.strip())
             msg += " parsing "  + "/".join(titles)
@@ -298,7 +326,7 @@ class Wtp(object):
     def _save_value(self, kind, args, nowiki):
         """Saves a value of a particular kind and returns a unique magic
         cookie for it."""
-        assert kind in ("T", "A", "L")  # Template/parserfn, arg, link
+        assert kind in ("T", "A", "L", "E")  # Template/parserfn, arg, link, ext
         assert isinstance(args, (list, tuple))
         assert nowiki in (True, False)
         # print("save_value", kind, args, nowiki)
@@ -320,11 +348,17 @@ class Wtp(object):
         """Encode all templates, template arguments, and parser function calls
         in the text, from innermost to outermost."""
 
+        def vbar_split(v):
+            args = list(m.group(1) for m in re.finditer(
+                r"(?si)\|((<\s*([-a-zA-z0-9]+)\b[^>]*>[^][{}]*?<\s*/\s*\3\s*>|"
+                r"[^|])*)", "|" + v))
+            return args
+
         def repl_arg(m):
             """Replacement function for template arguments."""
             nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
             orig = m.group(1)
-            args = orig.split("|")
+            args = vbar_split(orig)
             return self._save_value("A", args, nowiki)
 
         def repl_arg_err(m):
@@ -332,8 +366,8 @@ class Wtp(object):
             nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
             prefix = m.group(1)
             orig = m.group(2)
-            args = orig.split("|")
-            self.warning("heuristically added missing }} to template arg {}"
+            args = vbar_split(orig)
+            self.debug("heuristically added missing }} to template arg {}"
                          .format(args[0].strip()))
             return prefix + self._save_value("A", args, nowiki)
 
@@ -341,7 +375,8 @@ class Wtp(object):
             """Replacement function for templates {{name|...}} and parser
             functions."""
             nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
-            args = m.group(1).split("|")
+            v = m.group(1)
+            args = vbar_split(v)
             return self._save_value("T", args, nowiki)
 
         def repl_templ_err(m):
@@ -349,8 +384,9 @@ class Wtp(object):
             functions, with error."""
             nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
             prefix = m.group(1)
-            args = m.group(2).split("|")
-            self.warning("heuristically added missing }} to template {}"
+            v = m.group(2)
+            args = vbar_split(v)
+            self.debug("heuristically added missing }} to template {}"
                          .format(args[0].strip()))
             return prefix + self._save_value("T", args, nowiki)
 
@@ -358,9 +394,19 @@ class Wtp(object):
             """Replacement function for links [[...]]."""
             nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
             orig = m.group(1)
-            args = m.group(1).split("|")
+            args = vbar_split(orig)
             return self._save_value("L", args, nowiki)
 
+        def repl_extlink(m):
+            """Replacement function for external links [...].  This is also
+            used to replace bracketed sections, such as [...]."""
+            nowiki = m.group(0).find(MAGIC_NOWIKI_CHAR) >= 0
+            orig = m.group(1)
+            # Extlinks usually separate args by spaces, but this should do no
+            # harm here.
+            args = vbar_split(orig)
+            return self._save_value("E", args, nowiki)
+
         # As a preprocessing step, remove comments from the text
         text = re.sub(r"(?s)<!\s*--.*?--\s*>", "", text)
 
@@ -375,12 +421,16 @@ class Wtp(object):
             while True:
                 prev2 = text
                 # Encode links.
-                text = re.sub(r"\[" + MAGIC_NOWIKI_CHAR +
-                              r"?\[(([^][{}]|\[[^][{}]*\])+)\]" +
-                              # XXXremove: r"?\[([^][{}]*)\]" +
-                              # XXXremove: r"?\[(([^][{}]|\[[^]]*\])+)\]" +
-                              MAGIC_NOWIKI_CHAR + r"?\]",
-                              repl_link, text)
+                while True:
+                    text = re.sub(r"(?s)\[" + MAGIC_NOWIKI_CHAR +
+                                  r"?\[([^][{}]+)\]" +
+                                  MAGIC_NOWIKI_CHAR + r"?\]",
+                                  repl_link, text)
+                    if text == prev2:
+                        break
+                    prev2 = text
+                # Encode external links.
+                text = re.sub(r"(?s)\[([^][{}<>]+)\]", repl_extlink, text)
                 # Encode template arguments
                 text = re.sub(r"(?s)\{" + MAGIC_NOWIKI_CHAR +
                               r"?\{" + MAGIC_NOWIKI_CHAR +
@@ -404,9 +454,10 @@ class Wtp(object):
                     if text != prev2:
                         continue
                     break
-            # Encode templates
-            text = re.sub(r"(?s)\{" + MAGIC_NOWIKI_CHAR +
-                          r"?\{(([^{}]|\{\|[^{}]*\|\}|\}[^{}])+?)\}" +
+            text = re.sub(r"(?si)\{" + MAGIC_NOWIKI_CHAR +
+                          r"?\{(("
+                          r"[^{}]|\{\|[^{}]*\|\}|\}[^{}]|"
+                          r"[^{}][{}][^{}])+?)\}" +
                           MAGIC_NOWIKI_CHAR + r"?\}",
                           repl_templ, text)
             # We keep looping until there is no change during the iteration
@@ -421,6 +472,9 @@ class Wtp(object):
                               repl_templ_err, text)
                 if text != prev:
                     continue
+                # Replace remaining brackets and braces by corresponding
+                # character entities
+                # XXX
                 break
             prev = text
         # Replace any remaining braces etc by corresponding character entities
@@ -456,17 +510,6 @@ class Wtp(object):
         # itself but is included in transclusion.  Also text outside these tags
         # is included in transclusion.
         text = re.sub(r"(?is)<\s*(/\s*)?includeonly\s*(/\s*)?>", "", text)
-        # Sanity checks for certain unbalanced tags.  However, it
-        # appears some templates intentionally produce these and
-        # intend them to be displayed.  Thus don't warn, and we may
-        # even need to arrange for them to be properly parsed as text.
-        if False:
-           m = re.search(r"(?is)<\s*(/\s*)?noinclude\s*(/\s*)?>", text)
-           if m:
-               self.warning("unbalanced {}".format(m.group(0)))
-           m = re.search(r"(?is)<\s*(/\s*)?onlyinclude\s*(/\s*)?>", text)
-           if m:
-               self.warning("unbalanced {}".format(m.group(0)))
         return text
 
     def add_page(self, model, title, text, transient=False):
@@ -718,8 +761,6 @@ class Wtp(object):
         self.warnings, and self.debugs lists and any current section
         or subsection."""
         assert isinstance(title, str)
-        # variables and thus must be reloaded for each page.
-        self.lua = None  # Force reloading modules for every page
         self.title = title
         self.errors = []
         self.warnings = []
@@ -768,10 +809,17 @@ class Wtp(object):
             return "&lsqb;&lsqb;" + "&vert;".join(args) + "&rsqb;&rsqb;"
         return "[[" + "|".join(args) + "]]"
 
+    def _unexpanded_extlink(self, args, nowiki):
+        """Formats an unexpanded external link."""
+        if nowiki:
+            return "&lsqb;" + "&vert;".join(args) + "&rsqb;"
+        return "[" + "|".join(args) + "]"
+
     def expand(self, text, parent=None, pre_expand=False,
                template_fn=None, post_template_fn=None,
                templates_to_expand=None,
-               expand_parserfns=True, expand_invoke=True, quiet=False):
+               expand_parserfns=True, expand_invoke=True, quiet=False,
+               timeout=None):
         """Expands templates and parser functions (and optionally Lua macros)
         from ``text`` (which is from page with title ``title``).
         ``templates_to_expand`` should be None to expand all
@@ -797,6 +845,7 @@ class Wtp(object):
         assert isinstance(templates_to_expand, (set, dict, type(None)))
         assert self.title is not None  # start_page() must have been called
         assert quiet in (False, True)
+        assert timeout is None or isinstance(timeout, (int, float))
 
         # Handle <nowiki> in a preprocessing step
         text = preprocess_text(text)
@@ -833,13 +882,15 @@ class Wtp(object):
             assert isinstance(invoke_args, (list, tuple))
             assert callable(expander)
             assert isinstance(parent, (tuple, type(None)))
-            # print("invoke_fn", invoke_args)
+            # print("INVOKE_FN", invoke_args, parent)
             # sys.stdout.flush()
 
             # Use the Lua sandbox to execute a Lua macro.  This will initialize
             # the Lua environment and store it in self.lua if it does not
             # already exist (it needs to be re-created for each new page).
-            ret = call_lua_sandbox(self, invoke_args, expander, parent)
+            ret = call_lua_sandbox(self, invoke_args, expander, parent, timeout)
+            # print("invoke_fn: invoke_args={} parent={} LUA ret={!r}"
+            #       .format(invoke_args, parent, ret))
             return ret
 
         def expand_recurse(coded, parent, templates_to_expand):
@@ -880,9 +931,9 @@ class Wtp(object):
                     if kind == "A":
                         # Template argument reference
                         if len(args) > 2:
-                            self.warning("too many args ({}) in argument "
-                                         "reference {!r}"
-                                         .format(len(args), args))
+                            self.debug("too many args ({}) in argument "
+                                       "reference: {!r}"
+                                       .format(len(args), args))
                         self.expand_stack.append("ARG-NAME")
                         k = expand_recurse(expand_args(args[0], argmap),
                                            parent, all_templates).strip()
@@ -909,6 +960,12 @@ class Wtp(object):
                                         for x in args)
                         parts.append(self._unexpanded_link(new_args, nowiki))
                         continue
+                    if kind == "E":
+                        # Link to another page
+                        new_args = list(expand_args(x, argmap)
+                                        for x in args)
+                        parts.append(self._unexpanded_extlink(new_args, nowiki))
+                        continue
                     self.error("expand_arg: unsupported cookie kind {!r} in {}"
                                .format(kind, m.group(0)))
                     parts.append(m.group(0))
@@ -1010,8 +1067,7 @@ class Wtp(object):
                     # Check for undefined templates
                     if name not in all_templates:
                         # XXX tons of these in enwiktionary-20201201 ???
-                        #self.warning("undefined template {!r}"
-                        #             .format(tname))
+                        #self.debug("undefined template {!r}.format(tname))
                         parts.append('<strong class="error">Template:{}'
                                      '</strong>'
                                      .format(html.escape(name)))
@@ -1038,7 +1094,7 @@ class Wtp(object):
                     num = 1
                     for i in range(1, len(args)):
                         arg = str(args[i])
-                        m = re.match(r"""(?s)^\s*([^<>="']+?)\s*="""
+                        m = re.match(r"""(?s)^\s*([^][&<>="'\s]+?)\s*="""
                                      """\s*(.*?)\s*$""",
                                      arg)
                         if m:
@@ -1050,8 +1106,9 @@ class Wtp(object):
                             if k.isdigit():
                                 k = int(k)
                                 if k < 1 or k > 1000:
-                                    ctx.error("invalid argument number {}"
-                                              .format(k))
+                                    self.debug("invalid argument number {} "
+                                               "for template {!r}"
+                                               .format(k, name))
                                     k = 1000
                                 if num <= k:
                                     num = k + 1
@@ -1074,8 +1131,11 @@ class Wtp(object):
                     # Expand the body, either using ``template_fn`` or using
                     # normal template expansion
                     t = None
+                    # print("EXPANDING TEMPLATE: {} {}".format(name, ht))
                     if template_fn is not None:
                         t = template_fn(urllib.parse.unquote(name), ht)
+                        # print("TEMPLATE_FN {}: {} {} -> {}"
+                        #      .format(template_fn, name, ht, repr(t)))
                     if t is None:
                         if name in self.transient_templates:
                             body = self.transient_templates[name]
@@ -1118,6 +1178,8 @@ class Wtp(object):
 
                     # If a post_template_fn has been supplied, call it now
                     # to capture or alter the expansion
+                    # print("TEMPLATE EXPANDED: {} {} -> {!r}"
+                    #       .format(name, ht, t))
                     if post_template_fn is not None:
                         t2 = post_template_fn(urllib.parse.unquote(name), ht, t)
                         if t2 is not None:
@@ -1139,6 +1201,17 @@ class Wtp(object):
                                         for x in args)
                         self.expand_stack.pop()
                         parts.append(self._unexpanded_link(new_args, nowiki))
+                elif kind == "E":
+                    if nowiki:
+                        parts.append(self._unexpanded_extlink(args, nowiki))
+                    else:
+                        # Link to an external page
+                        self.expand_stack.append("[extlink]")
+                        new_args = list(expand_recurse(x, parent,
+                                                       templates_to_expand)
+                                        for x in args)
+                        self.expand_stack.pop()
+                        parts.append(self._unexpanded_extlink(new_args, nowiki))
                 else:
                     self.error("expand: unsupported cookie kind {!r} in {}"
                                .format(kind, m.group(0)))
@@ -1155,10 +1228,10 @@ class Wtp(object):
         expanded = expand_recurse(encoded, parent, templates_to_expand)
 
         # Expand any remaining magic cookies and remove nowiki char
-        expanded = self._finalize_expand(expanded, False)
+        expanded = self._finalize_expand(expanded)
         return expanded
 
-    def _finalize_expand(self, text, unescape):
+    def _finalize_expand(self, text):
         """Expands any remaining magic characters (to their original values)
         and removes nowiki characters."""
 
@@ -1171,6 +1244,8 @@ class Wtp(object):
                 return self._unexpanded_arg(args, nowiki)
             if kind == "L":
                 return self._unexpanded_link(args, nowiki)
+            if kind == "E":
+                return self._unexpanded_extlink(args, nowiki)
             self.error("magic_repl: unsupported cookie kind {!r}"
                        .format(kind))
             return ""
@@ -1184,15 +1259,9 @@ class Wtp(object):
             if prev == text:
                 break
 
-        # Unescape HTML entities if so requested (we don't do this at the end
-        # of normal expansion, but we do it at the end of parsing)
-        if unescape:
-            text = html.unescape(text)
-            text = re.sub(MAGIC_NOWIKI_CHAR, "", text)
-        else:
-            # Convert the special <nowiki /> character back to <nowiki />.
-            # This is done at the end of normal expansion.
-            text = re.sub(MAGIC_NOWIKI_CHAR, "<nowiki />", text)
+        # Convert the special <nowiki /> character back to <nowiki />.
+        # This is done at the end of normal expansion.
+        text = re.sub(MAGIC_NOWIKI_CHAR, "<nowiki />", text)
         return text
 
     def process(self, path, page_handler, phase1_only=False):
@@ -1214,7 +1283,9 @@ class Wtp(object):
         assert isinstance(path, str)
         assert callable(page_handler)
         # Process the dump and copy it to temporary file (Phase 1)
-        process_dump(self, path, page_handler, phase1_only)
+        process_dump(self, path, page_handler)
+        if phase1_only:
+            return []
 
         # Reprocess all the pages that we captured in Phase 1
         return self.reprocess(page_handler)
@@ -1243,9 +1314,15 @@ class Wtp(object):
             # Single-threaded version (without subprocessing).  This is
             # primarily intended for debugging.
             for model, title in self.page_seq:
-                success, ret = phase2_page_handler((model, title))
+                success, ret_title, t, ret = phase2_page_handler((model, title))
+                assert ret_title == title
                 if not success:
                     print(ret)  # Print error in parent process - do not remove
+                    lines = ret.split("\n")
+                    msg = lines[0]
+                    trace = "\n".join(lines[1:])
+                    if msg.find("EXCEPTION") >= 0:
+                        self.error(msg, trace=trace)
                     continue
                 if ret is not None:
                     yield ret
@@ -1257,26 +1334,48 @@ class Wtp(object):
             else:
                 pool = multiprocessing.Pool(self.num_threads)
             cnt = 0
-            for success, ret in pool.imap_unordered(phase2_page_handler,
-                                                    self.page_seq, 64):
+            start_t = time.time()
+            last_t = time.time()
+            for success, title, t, ret in \
+                pool.imap_unordered(phase2_page_handler, self.page_seq, 64):
+                if t + 300 < time.time():
+                    print("====== REPROCESS GOT OLD RESULT ({:.1f}s): {}"
+                          .format(time.time() - t, title))
+                sys.stdout.flush()
                 if not success:
-                    print(ret)  # Print error in parent process - do not remove
+                    # Print error in parent process - do not remove
+                    print(ret)
+                    sys.stdout.flush()
                     continue
                 if ret is not None:
                     yield ret
                 cnt += 1
-                if not self.quiet and cnt % 1000 == 0:
-                    print("  ... {}/{} pages ({:.1%}) processed"
+                if (not self.quiet and
+                    # cnt % 1000 == 0 and
+                    time.time() - last_t > 1):
+                    remaining = len(self.page_seq) - cnt
+                    secs = (time.time() - start_t) / cnt * remaining
+                    print("  ... {}/{} pages ({:.1%}) processed, "
+                          "{:02d}:{:02d}:{:02d} remaining"
                           .format(cnt, len(self.page_seq),
-                                  cnt / len(self.page_seq)))
+                                  cnt / len(self.page_seq),
+                                  int(secs / 3600),
+                                  int(secs / 60 % 60),
+                                  int(secs % 60)))
                     sys.stdout.flush()
+                    last_t = time.time()
             pool.close()
             pool.join()
 
+        sys.stderr.flush()
+        sys.stdout.flush()
+
     def page_exists(self, title):
         """Returns True if the given page exists, and False if it does not
         exist."""
         assert isinstance(title, str)
+        if title.startswith("Main:"):
+            title = title[5:]
         # XXX should we canonicalize title?
         if title in self.transient_pages:
             return True
@@ -1286,6 +1385,8 @@ class Wtp(object):
         """Reads the contents of the page.  Returns None if the page does
         not exist."""
         assert isinstance(title, str)
+        if title.startswith("Main:"):
+            title = title[5:]
         # XXX should we canonicalize title?
         if title in self.transient_pages:
             title, model, text = self.transient_pages[title]
@@ -1301,7 +1402,7 @@ class Wtp(object):
         return rawdata.decode("utf-8")
 
     def parse(self, text, pre_expand=False, expand_all=False,
-              additional_expand=None):
+              additional_expand=None, template_fn=None, post_template_fn=None):
         """Parses the given text into a parse tree (WikiNode tree).  If
         ``pre_expand`` is True, then before parsing this will expand
         those templates that have been detected to potentially
@@ -1322,10 +1423,13 @@ class Wtp(object):
 
         # Expand some or all templates in the text as requested
         if expand_all:
-            text = self.expand(text)
+            text = self.expand(text, template_fn=template_fn,
+                               post_template_fn=post_template_fn)
         elif pre_expand or additional_expand:
             text = self.expand(text, pre_expand=pre_expand,
-                               templates_to_expand=additional_expand)
+                               templates_to_expand=additional_expand,
+                               template_fn=template_fn,
+                               post_template_fn=post_template_fn)
 
         # print("parse:", repr(text))
 
@@ -1345,7 +1449,8 @@ class Wtp(object):
 
     def node_to_wikitext(self, node):
         """Converts the given parse tree node back to Wikitext."""
-        return to_wikitext(node)
+        v = to_wikitext(node)
+        return v
 
     def node_to_html(self, node, template_fn=None, post_template_fn=None):
         """Converts the given parse tree node to HTML."""
