@@ -1,20 +1,18 @@
 -- Simplified implementation of mw.hash for running WikiMedia Scribunto code
 -- under Python
 --
--- Copyright (c) 2020 Tatu Ylonen.  See file LICENSE and https://ylonen.org
+-- Copyright (c) 2020-2021 Tatu Ylonen.  See file LICENSE and https://ylonen.org
 
-function mw_hash_hashValue(algo, value)
+local mw_hash = {
+}
+
+function mw_hash.hashValue(algo, value)
   print("MW_HASH_HASHVALUE")
 end
 
-function mw_hash_listAlgorithms()
+function mw_hash.listAlgorithms()
   print("MW_HASH_LISTALGORITHMS")
   return {}
 end
 
-local mw_hash = {
-  hashValue = mw_hash_hashValue,
-  listAlgorithms = mw_hash_listAlgorithms,
-}
-
 return mw_hash
