@@ -10,6 +10,7 @@ import json
 import traceback
 import unicodedata
 import pkg_resources
+#from lru import LRU
 import lupa
 from lupa import LuaRuntime
 from .parserfns import PARSER_FUNCTIONS, call_parser_function, tag_fn
@@ -34,50 +35,107 @@ LANGUAGE_CODE_TO_NAME = { x["code"]: x["name"]
                           for x in ALL_LANGUAGES
                           if x.get("code") and x.get("name") }
 
+# Substitutions to perform on Lua code from the dump to convert from
+# Lua 5.1 to current 5.3
+loader_replace_patterns = list((re.compile(src), dst) for src, dst in [
+    [r"\\\\\?", r"%\\092?"],
+    [r"\\\\\*", r"%\\092*"],
+    [r"\\\\\-", r"%\\092-"],
+    [r"\\\\\+", r"%\\092+"],
+    [r"\\\\\|", r"%\\092|"],
+    [r"\\\\\[", r"%\\092["],
+    # [r"'\\\\'", r"'\092'"],  # now covered by the one below
+    [r"\\\\", r"\\092"],
+
+    [r"\\\[", r"%%["],
+    [r"\\:", r":"],
+    [r"\\,", r","],
+    [r"\\\(", r"%("],
+    [r"\\\)", r"%)"],
+    [r"\\\+", r"%+"],
+    [r"\\\*", r"%*"],
+    [r"\\>", r">"],
+    [r"\\\.", r"%."],
+    [r"\\\?", r"%?"],
+    [r"\\-", r"%-"],
+    [r"\\!", r"!"],
+    [r"\\\|", r"|"],
+    [r"\\\^", r"%^"],
+    [r"\\ʺ", r"ʺ"],
+    [r"\\s", r"%s"],
+
+    # Workaround kludge for a bug in Lua 5.4 (lupa 1.10)
+    [r"\[(\w+)\s*==\s*true\]", r"[not not \1]"],
+])
+
+# -- Wikimedia uses an older version of Lua.  Make certain substitutions
+# -- to make existing code run on more modern versions of Lua.
+# content = string.gsub(content, "\\\\", "\\092")
+# content = string.gsub(content, "%%\\%[", "%%%%[")
+# content = string.gsub(content, "\\:", ":")
+# content = string.gsub(content, "\\,", ",")
+# content = string.gsub(content, "\\%(", "%%(")
+# content = string.gsub(content, "\\%)", "%%)")
+# content = string.gsub(content, "\\%+", "%%+")
+# content = string.gsub(content, "\\%*", "%%*")
+# content = string.gsub(content, "\\>", ">")
+# content = string.gsub(content, "\\%.", "%%.")
+# content = string.gsub(content, "\\%?", "%%?")
+# content = string.gsub(content, "\\%-", "%%-")
+# content = string.gsub(content, "\\!", "!")
+# content = string.gsub(content, "\\|", "|")  -- XXX tentative, see ryu:951
+# content = string.gsub(content, "\\ʺ", "ʺ")
+
 def lua_loader(ctx, modname):
     """This function is called from the Lua sandbox to load a Lua module.
     This will load it from either the user-defined modules on special
     pages or from a built-in module in the file system.  This returns None
     if the module could not be loaded."""
-    # print("lua_loader", modname)
+    # print("LUA_LOADER IN PYTHON:", modname)
     assert isinstance(modname, str)
-    # XXX consider implementing some kind of cache here, e.g., LRU cache
-    # for the most recently used modules (modname -> text)
-    # print("Loading", modname)
     modname = modname.strip()
     if modname.startswith("Module:"):
-        modname = modname[7:]
-    modname1 = "Module:" + modname
-    modname1 = ctx._canonicalize_template_name(modname1)
-    body = ctx.read_by_title(modname1)
-    if body is not None:
-        return body
-    path = modname
-    path = re.sub(r":", "/", path)
-    path = re.sub(r" ", "_", path)
-    # path = re.sub(r"\.", "/", path)
-    path = re.sub(r"//+", "/", path)
-    path = re.sub(r"\.\.", ".", path)
-    if path.startswith("/"):
-        path = path[1:]
-    path += ".lua"
-    for prefix, exceptions in builtin_lua_search_paths:
-        if modname in exceptions:
-            continue
-        p = lua_dir + prefix + "/" + path
-        if os.path.isfile(p):
-            with open(p, "r") as f:
-                data = f.read()
-            return data
-    # It is very common that modules are not found; e.g., many Chinese macros
-    # attempt to load modules related to the characters and generate tons
-    # of these errors.
-    #ctx.error("Lua module not found: {}"
-    #          .format(modname))
-    return None
+        # Canonicalize the name
+        modname = ctx._canonicalize_template_name(modname)
 
+        # First try to load it as a module
+        if modname.startswith("Module:_"):
+            # Module names starting with _ are considered internal and cannot be
+            # loaded from the dump file for security reasons.  This is to ensure
+            # that the sandbox always gets loaded from a local file.
+            data = None
+        else:
+            data = ctx.read_by_title(modname)
+    else:
+        # Try to load it from a file
+        path = modname
+        path = re.sub(r"[\0-\037]", "", path)  # Remove control chars, e.g. \n
+        path = re.sub(r":", "/", path)      # Replace : by /
+        path = re.sub(r" ", "_", path)      # Replace spaces by underscore
+        path = re.sub(r"//+", "/", path)    # Replace multiple slashes by one
+        path = re.sub(r"\.\.+", ".", path)  # Replace .. and longer by .
+        path = re.sub(r"^//+", "", path)    # Remove initial slashes
+        path += ".lua"
+        data = None
+        for prefix, exceptions in builtin_lua_search_paths:
+            if modname in exceptions:
+                continue
+            p = lua_dir + prefix + "/" + path
+            if os.path.isfile(p):
+                with open(p, "r", encoding="utf-8") as f:
+                    data = f.read()
+                break
 
-def mw_text_decode(text, decodeNamedEntities=False):
+    if data is None:
+        # We did not find the module
+        return None
+
+    # Perform compatibility substitutions on the Lua code
+    for src, dst in loader_replace_patterns:
+        data = re.sub(src, dst, data)
+    return data
+
+def mw_text_decode(text, decodeNamedEntities):
     """Implements the mw.text.decode function for Lua code."""
     if decodeNamedEntities:
         return html.unescape(text)
@@ -105,7 +163,7 @@ def mw_text_decode(text, decodeNamedEntities=False):
     parts.append(text[pos:])
     return "".join(parts)
 
-def mw_text_encode(text, charset='<>&\xa0"'):
+def mw_text_encode(text, charset):
     """Implements the mw.text.encode function for Lua code."""
     parts = []
     for ch in str(text):
@@ -188,9 +246,8 @@ def get_page_info(ctx, title):
     titles."""
     assert isinstance(title, str)
 
-    # XXX actually look at information collected in phase 1 to determine
     page_id = 0  # XXX collect required info in phase 1
-    page_exists = False  # XXX collect required info in Phase 1
+    page_exists = ctx.page_exists(title)
     redirect_to = ctx.redirects.get(title, None)
 
     # whether the page exists and what its id might be
@@ -213,9 +270,6 @@ def get_page_content(ctx, title):
     # Read the page by its title
     data = ctx.read_by_title(title)
     if data is None:
-        # ctx.warning("attempted to access page content for {!r} which "
-        #             "is not available"
-        #             .format(title))
         return None
     return data
 
@@ -238,10 +292,20 @@ def fetch_language_names(ctx, include):
     return ctx.lua.table_from(dt)
 
 
+def call_set_functions(ctx, set_functions):
+    # Set functions that are implemented in Python
+    set_functions(mw_text_decode,
+                  mw_text_encode,
+                  mw_text_jsonencode,
+                  lambda x, *rest:
+                  mw_text_jsondecode(ctx, x, *rest),
+                  lambda x: get_page_info(ctx, x),
+                  lambda x: get_page_content(ctx, x),
+                  fetch_language_name,
+                  lambda x: fetch_language_names(ctx, x))
+
+
 def initialize_lua(ctx):
-    assert ctx.lua is None
-    # Load Lua sandbox code.
-    lua_sandbox = open(lua_dir + "sandbox.lua").read()
 
     def filter_attribute_access(obj, attr_name, is_setting):
         if isinstance(attr_name, unicode):
@@ -252,48 +316,71 @@ def initialize_lua(ctx):
     lua = LuaRuntime(unpack_returned_tuples=True,
                      register_eval=False,
                      attribute_filter=filter_attribute_access)
-    lua.execute(lua_sandbox)
-    lua.eval("lua_set_loader")(lambda x: lua_loader(ctx, x),
-                               mw_text_decode,
-                               mw_text_encode,
-                               mw_text_jsonencode,
-                               lambda x, *rest:
-                               mw_text_jsondecode(ctx, x, *rest),
-                               lambda x: get_page_info(ctx, x),
-                               lambda x: get_page_content(ctx, x),
-                               fetch_language_name,
-                               lambda x: fetch_language_names(ctx, x))
     ctx.lua = lua
 
+    # Load Lua sandbox Phase 1.  This is a very minimal file that only sets
+    # the Lua loader to our custom loader; we will then use it to load the
+    # bigger phase 2 of the sandbox.  This way, most of the sandbox loading
+    # will benefit from caching and precompilation (when implemented).
+    lua_sandbox = open(lua_dir + "_sandbox_phase1.lua", encoding="utf-8").read()
+    set_loader = lua.execute(lua_sandbox)
+    # Call the function that sets the Lua loader
+    set_loader(lambda x: lua_loader(ctx, x))
 
-def call_lua_sandbox(ctx, invoke_args, expander, parent):
-    """Calls a function in a Lua module in the Lua sandbox.  This creates
-    the sandbox instance if it does not already exist and stores it in
-    ctx.lua.  ``invoke_args`` is the arguments to the call;
-    ``expander`` should be a function to expand an argument.
-    ``parent`` should be None or (parent_title, parent_args) for the
-    parent page."""
+    # Then load the second phase of the sandbox.  This now goes through the
+    # new loader and is evaluated in the sandbox.  This mostly implements
+    # compatibility code.
+    ret = lua.eval('new_require("_sandbox_phase2")')
+    set_functions = ret[1]
+    ctx.lua_invoke = ret[2]
+    ctx.lua_reset_env = ret[3]
+
+    # Set Python functions for Lua
+    call_set_functions(ctx, set_functions)
+
+
+def call_lua_sandbox(ctx, invoke_args, expander, parent, timeout):
+    """Calls a function in a Lua module in the Lua sandbox.
+    ``invoke_args`` is the arguments to the call; ``expander`` should
+    be a function to expand an argument.  ``parent`` should be None or
+    (parent_title, parent_args) for the parent page."""
     assert isinstance(invoke_args, (list, tuple))
     assert callable(expander)
     assert parent is None or isinstance(parent, (list, tuple))
+    assert timeout is None or isinstance(timeout, (int, float))
 
-    # print("call_lua_sandbox", invoke_args, parent)
+    # print("{}: CALL_LUA_SANDBOX: {} {}"
+    #       .format(ctx.title, invoke_args, parent))
 
     if len(invoke_args) < 2:
-        ctx.error("#invoke {}: too few arguments"
+        ctx.debug("#invoke {} with too few arguments"
                   .format(invoke_args))
         return ("{{" + invoke_args[0] + ":" +
                 "|".join(invoke_args[1:]) + "}}")
 
+    # Initialize the Lua sandbox if not already initialized
+    if ctx.lua_depth == 0:
+        if ctx.lua is None:
+            # This is the first call to the Lua sandbox.
+            # Create a Lua context and initialize it.
+            initialize_lua(ctx)  # This sets ctx.lua
+        else:
+            # This is a second or later call to the Lua sandbox.
+            # Reset the Lua context back to initial state.
+            ctx.lua_reset_env()
+            ret = ctx.lua.eval('new_require("_sandbox_phase2")')
+            set_functions = ret[1]
+            ctx.lua_invoke = ret[2]
+            ctx.lua_reset_env = ret[3]
+            call_set_functions(ctx, set_functions)
+
+    ctx.lua_depth += 1
+    lua = ctx.lua
+
     # Get module and function name
     modname = expander(invoke_args[0]).strip()
     modfn = expander(invoke_args[1]).strip()
 
-    # Initialize the Lua sandbox if not already initialized
-    if ctx.lua is None:
-        initialize_lua(ctx)
-    lua = ctx.lua
-
     def value_with_expand(frame, fexpander, x):
         assert isinstance(frame, dict)
         assert isinstance(fexpander, str)
@@ -313,8 +400,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
             frame_args = {}
             num = 1
             for arg in args:
-                m = re.match(r"""^\s*([^<>="']+?)\s*=\s*(.*?)\s*$""",
-                             arg)
+                m = re.match(r"""^(?s)\s*([^<>="']+?)\s*=\s*(.*?)\s*$""", arg)
                 if m:
                     # Have argument name
                     k, arg = m.groups()
@@ -334,7 +420,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
 
         def extensionTag(frame, *args):
             if len(args) < 1:
-                ctx.error("extensionTag: missing arguments")
+                ctx.debug("lua extensionTag with missing arguments")
                 return ""
             dt = args[0]
             if not isinstance(dt, (str, int, float, type(None))):
@@ -375,7 +461,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
 
         def callParserFunction(frame, *args):
             if len(args) < 1:
-                ctx.error("callParserFunction: missing name")
+                ctx.debug("lua callParserFunction missing name")
                 return ""
             name = args[0]
             if not isinstance(name, str):
@@ -397,7 +483,8 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
                         new_args.append(str(v))
             name = ctx._canonicalize_parserfn_name(name)
             if name not in PARSER_FUNCTIONS:
-                ctx.error("frame:callParserFunction(): undefined function {!r}"
+                ctx.debug("lua frame callParserFunction() undefined "
+                          "function {!r}"
                           .format(name))
                 return ""
             return call_parser_function(ctx, name, new_args, lambda x: x)
@@ -412,7 +499,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
 
         def preprocess(frame, *args):
             if len(args) < 1:
-                ctx.error("preprocess: missing argument")
+                ctx.debug("lua preprocess missing argument")
                 return ""
             v = args[0]
             if not isinstance(v, str):
@@ -427,11 +514,11 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
 
         def expandTemplate(frame, *args):
             if len(args) < 1:
-                ctx.error("expandTemplate: missing arguments")
+                ctx.debug("lua expandTemplate missing arguments")
                 return ""
             dt = args[0]
             if isinstance(dt, (int, float, str, type(None))):
-                ctx.error("expandTemplate: arguments should be named")
+                ctx.debug("lua expandTemplate arguments should be named")
                 return ""
             title = dt["title"] or ""
             args = dt["args"] or {}
@@ -482,7 +569,7 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
     stack_len = len(ctx.expand_stack)
     ctx.expand_stack.append("Lua:{}:{}()".format(modname, modfn))
     try:
-        ret = lua.eval("lua_invoke")(modname, modfn, frame, ctx.title)
+        ret = ctx.lua_invoke(modname, modfn, frame, ctx.title, timeout)
         if not isinstance(ret, (list, tuple)):
             ok, text = ret, ""
         elif len(ret) == 1:
@@ -490,15 +577,19 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
         else:
             ok, text = ret[0], ret[1]
     except UnicodeDecodeError:
-        ctx.error("invalid unicode returned by {} parent {}"
+        ctx.debug("invalid unicode returned from lua by {}: parent {}"
                   .format(invoke_args, parent))
         ok, text = True, ""
+    except lupa._lupa.LuaError as e:
+        ok, text = False, e
     finally:
         while len(ctx.expand_stack) > stack_len:
             ctx.expand_stack.pop()
+    # print("Lua call returned: {}".format(invoke_args))
+    ctx.lua_depth -= 1
     if ok:
         if text is None:
-            text = "nil"
+            text = ""
         text = str(text)
         text = unicodedata.normalize("NFC", text)
         return text
@@ -515,12 +606,16 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
     msg = re.sub(r".*?:\d+: ", "", text.split("\n")[0])
     if text.find("'debug.error'") >= 0:
         if not msg.startswith("This template is deprecated."):
-            ctx.warning(msg)
-    # XXX This probably should be deleted; it can cause important errors
-    # to be missed
-    # elif text.find("in function 'error'") >= 0:
-    #     if not msg.startswith('The parameter "'):
-    #         ctx.warning(msg)
+            ctx.debug("lua error -- " + msg)
+    elif text.find("Translations must be for attested and approved ") >= 0:
+        # Ignore this error - it is an error but a clear error in Wiktionary
+        # rather than in the extractor.
+        return ""
+    elif (text.find("attempt to index a nil value (local 'lang')") >= 0 and
+          text.find("in function 'Module:links.getLinkPage'") >= 0):
+        # Ignore this error - happens when an unknown language code is passed
+        # to various templates (a Wiktionary error, not extractor error)
+        return ""
     else:
         parts = []
         in_traceback = 0
@@ -530,9 +625,14 @@ def call_lua_sandbox(ctx, invoke_args, expander, parent):
             #    break
             parts.append(line)
         trace = "\n".join(parts)
-        ctx.error("LUA error in #invoke {} parent {}"
-                  .format(invoke_args, parent),
-                  trace=trace)
+        if "check deprecated lang param usage" in ctx.expand_stack:
+            ctx.debug("LUA error but likely not bug -- in #invoke {} parent {}"
+                      .format(invoke_args, parent),
+                      trace=trace)
+        else:
+            ctx.error("LUA error in #invoke {} parent {}"
+                      .format(invoke_args, parent),
+                      trace=trace)
     msg = "Lua execution error"
     if text.find("Lua timeout error") >= 0:
         msg = "Lua timeout error"
