@@ -1,6 +1,6 @@
 # Simple WikiMedia markup (WikiText) syntax parser
 #
-# Copyright (c) 2020 Tatu Ylonen.  See file LICENSE and https://ylonen.org
+# Copyright (c) 2020-2021 Tatu Ylonen.  See file LICENSE and https://ylonen.org
 
 import re
 import enum
@@ -277,13 +277,13 @@ def _parser_merge_str_children(ctx):
             strings.append(x)
         else:
             if strings:
-                s = ctx._finalize_expand("".join(strings), False)
+                s = ctx._finalize_expand("".join(strings))
                 if s:
                     new_children.append(s)
                 strings = []
             new_children.append(x)
     if strings:
-        s = ctx._finalize_expand("".join(strings), False)
+        s = ctx._finalize_expand("".join(strings))
         if s:
             new_children.append(s)
     node.children = new_children
@@ -302,16 +302,16 @@ def _parser_pop(ctx, warn_unclosed):
     # Warn about unclosed syntaxes.
     if warn_unclosed and node.kind in MUST_CLOSE_KINDS:
         if node.kind == NodeKind.HTML:
-            ctx.warning("HTML tag <{}> not properly closed".format(node.args),
+            ctx.debug("HTML tag <{}> not properly closed".format(node.args),
                         trace="started on line {}, detected on line {}"
                         .format(node.loc, ctx.linenum))
         elif node.kind == NodeKind.PARSER_FN:
-            ctx.warning("parser function {!r} not properly closed"
+            ctx.debug("parser function invocation {!r} not properly closed"
                         .format(node.args[0]),
                         trace="started on line {}, detected on line {}"
                         .format(node.loc, ctx.linenum))
         elif node.kind == NodeKind.URL and not node.children:
-            # This can happen at least when [ is inside tempate argument.
+            # This can happen at least when [ is inside template argument.
             ctx.parser_stack.pop()
             node2 = ctx.parser_stack[-1]
             node3 = node2.children.pop()
@@ -324,9 +324,9 @@ def _parser_pop(ctx, warn_unclosed):
             # them.
             pass
         else:
-            ctx.warning("{} not properly closed".format(node.kind.name),
-                        trace="started on line {}, detected on line {}"
-                        .format(node.loc, ctx.linenum))
+            ctx.debug("{} not properly closed".format(node.kind.name),
+                      trace="started on line {}, detected on line {}"
+                      .format(node.loc, ctx.linenum))
 
     # When popping BOLD and ITALIC nodes, if the node has no children,
     # just remove the node from it's parent's children.  We may otherwise
@@ -377,19 +377,31 @@ def _parser_have(ctx, kind):
     return False
 
 
+def close_begline_lists(ctx):
+    """Closes currently open list if at the beginning of a line."""
+    if not ctx.beginning_of_line:
+        return
+    node = ctx.parser_stack[-1]
+    while node.kind in (NodeKind.LIST, NodeKind.LIST_ITEM):
+        _parser_pop(ctx, False)
+        node = ctx.parser_stack[-1]
+
+
 def text_fn(ctx, token):
     """Inserts the token as raw text into the parse tree."""
     node = ctx.parser_stack[-1]
 
     # Convert certain characters from the token into HTML entities
-    token = re.sub(r"<", "&lt;", token)
-    token = re.sub(r">", "&gt;", token)
+    # XXX this breaks tags inside templates, e.g. <math> in
+    # "conjugacy class"/English examples
+    #token = re.sub(r"<", "&lt;", token)
+    #token = re.sub(r">", "&gt;", token)
 
     # External links [https://...] require some magic.  They only seem to
     # be links if the content looks like a URL."""
     if node.kind == NodeKind.URL:
         if not node.args and not node.children:
-            if not re.match(r"^(http|https|mailto):", token):
+            if not re.match(r"^(https?:|mailto:|//)", token):
                 # It does not look like a URL
                 ctx.parser_stack.pop()
                 node2 = ctx.parser_stack[-1]
@@ -432,13 +444,16 @@ def text_fn(ctx, token):
                     continue
             elif node.kind in (NodeKind.BOLD, NodeKind.ITALIC):
                 _parser_merge_str_children(ctx)
-                ctx.warning("{} not properly closed on the same line"
-                            .format(node.kind.name))
+                ctx.debug("{} not properly closed on the same line"
+                          .format(node.kind.name))
                 _parser_pop(ctx, False)
             break
 
         # Spaces at the beginning of a line indicate preformatted text
         if token.startswith(" ") or token.startswith("\t"):
+            if ctx.parser_stack[-1].kind in (NodeKind.TABLE,
+                                             NodeKind.TABLE_ROW):
+                return
             if node.kind != NodeKind.PREFORMATTED and not ctx.pre_parse:
                 node = _parser_push(ctx, NodeKind.PREFORMATTED)
 
@@ -522,7 +537,7 @@ def subtitle_end_fn(ctx, token):
     # Move children of the subtitle node to be its first argument.
     node = ctx.parser_stack[-1]
     if node.kind != kind:
-        ctx.warning("subtitle start and end markers level mismatch")
+        ctx.debug("subtitle start and end markers level mismatch")
     _parser_merge_str_children(ctx)
     node.args.append(node.children)
     node.children = []
@@ -590,14 +605,6 @@ def bold_fn(ctx, token):
         _parser_push(ctx, NodeKind.ITALIC)
 
 
-def ilink_start_fn(ctx, token):
-    """Processes an internal link start token "[["."""
-    if ctx.pre_parse:
-        return text_fn(ctx, token)
-
-    _parser_push(ctx, NodeKind.LINK)
-
-
 def elink_start_fn(ctx, token):
     """Processes an external link start token "["."""
     if ctx.pre_parse:
@@ -618,6 +625,9 @@ def elink_end_fn(ctx, token):
         if node.kind == NodeKind.URL:
             _parser_pop(ctx, False)
             break
+        if node.kind in (NodeKind.TEMPLATE, NodeKind.TEMPLATE_ARG,
+                         NodeKind.LINK, NodeKind.ITALIC, NodeKind.BOLD):
+            return text_fn(ctx, token)
         _parser_pop(ctx, True)
 
 
@@ -647,8 +657,12 @@ def url_fn(ctx, token):
 def magic_fn(ctx, token):
     """Handler for a magic character used to encode templates, template
     arguments, and parser function calls."""
+    # Close lists if at the beginning of a line
+    close_begline_lists(ctx)
+    # Handle the magic character token
     idx = ord(token) - MAGIC_FIRST
     kind, args, nowiki = ctx.cookies[idx]
+    # print("MAGIC_FN:", kind, args, nowiki)
     ctx.beginning_of_line = False
     if kind == "T":
         if nowiki:
@@ -700,7 +714,8 @@ def magic_fn(ctx, token):
 
     elif kind == "L":
         if nowiki:
-            process_text(ctx, "&lsqb;&lsqb;" + args[0] + "&rsqb;&rsqb;")
+            process_text(ctx, "&lsqb;&lsqb;" + "&vert;".join(args) +
+                         "&rsqb;&rsqb;")
             return
         # Link to another page
         _parser_push(ctx, NodeKind.LINK)
@@ -719,6 +734,37 @@ def magic_fn(ctx, token):
                 _parser_pop(ctx, False)
                 break
             _parser_pop(ctx, True)
+
+    elif kind == "E":
+        # Link to an external page (or just text in brackets, e.g. [...])
+        if (not nowiki and args and
+            (args[0].find(":") >= 0 or args[0].startswith("//"))):
+            _parser_push(ctx, NodeKind.URL)
+
+            # Process arguments
+            process_text(ctx, args[0])
+            for arg in args[1:]:
+                vbar_fn(ctx, "|")
+                process_text(ctx, arg)
+
+            # The URL could have been popped if the content does not look like
+            # a URL.
+            if not _parser_have(ctx, NodeKind.URL):
+                # It must have been popped.
+                text_fn(ctx, "]")
+            else:
+                # Pop until we are back at this level and close the URL node
+                while True:
+                    node = ctx.parser_stack[-1]
+                    if node.kind == NodeKind.ROOT:
+                        break
+                    if node.kind == NodeKind.URL:
+                        _parser_pop(ctx, False)
+                        break
+                    _parser_pop(ctx, True)
+        else:
+            process_text(ctx, "[" + "&vert;".join(args) + "]")
+            return
     else:
         self.error("magic_fn: unsupported cookie kind {!r}"
                    .format(kind))
@@ -757,6 +803,7 @@ def table_start_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    close_begline_lists(ctx)
     _parser_push(ctx, NodeKind.TABLE)
 
 
@@ -774,6 +821,7 @@ def table_check_attrs(ctx):
 
 def table_row_check_attrs(ctx):
     """Checks if the table row has attributes, and if so, parses them."""
+    close_begline_lists(ctx)
     node = ctx.parser_stack[-1]
     if node.kind != NodeKind.TABLE_ROW:
         return
@@ -789,6 +837,7 @@ def table_caption_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    close_begline_lists(ctx)
     table_check_attrs(ctx)
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
@@ -805,13 +854,14 @@ def table_hdr_cell_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
-    if token == "!" and not ctx.beginning_of_line:
-        return text_fn(ctx, token)
-
+    close_begline_lists(ctx)
     table_row_check_attrs(ctx)
     table_check_attrs(ctx)
+
+    # Outside tables, just interpret ! and !! as raw text
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
+
     while True:
         node = ctx.parser_stack[-1]
         if node.kind == NodeKind.TABLE_ROW:
@@ -829,7 +879,14 @@ def table_hdr_cell_fn(ctx, token):
             else:
                 text_fn(ctx, token)
             return
-        if node.kind == NodeKind.TABLE_CELL:
+        if node.kind in (NodeKind.HTML, NodeKind.TEMPLATE):
+            # Inside nested HTML, interpret ! and !! as normal text
+            return text_fn(ctx, token)
+        if (node.kind == NodeKind.TABLE_CELL and
+            not ctx.beginning_of_line and
+            not ctx.wsp_beginning_of_line):
+            # Inside a cell, ! and !! are normal text unless at the beginning
+            # of a line
             return text_fn(ctx, token)
         _parser_pop(ctx, True)
 
@@ -839,6 +896,7 @@ def table_row_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    close_begline_lists(ctx)
     table_check_attrs(ctx)
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
@@ -855,13 +913,16 @@ def table_cell_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    close_begline_lists(ctx)
     table_row_check_attrs(ctx)
     table_check_attrs(ctx)
 
     if not _parser_have(ctx, NodeKind.TABLE):
         return text_fn(ctx, token)
 
-    if token == "|" and not ctx.beginning_of_line:
+    if (token == "|" and
+        not ctx.wsp_beginning_of_line and
+        not ctx.beginning_of_line):
         # This might separate attributes for captions, header cells, and
         # data cells
         _parser_merge_str_children(ctx)
@@ -885,6 +946,9 @@ def table_cell_fn(ctx, token):
             break
         if node.kind == NodeKind.TABLE_CAPTION:
             return text_fn(ctx, token)
+        if node.kind == NodeKind.HTML:
+            # Inside nested HTML, treat | and || as normal text
+            return text_fn(ctx, token)
         _parser_pop(ctx, True)
     _parser_push(ctx, NodeKind.TABLE_CELL)
 
@@ -922,6 +986,7 @@ def table_end_fn(ctx, token):
     if ctx.pre_parse:
         return text_fn(ctx, token)
 
+    close_begline_lists(ctx)
     table_row_check_attrs(ctx)
     table_check_attrs(ctx)
     if not _parser_have(ctx, NodeKind.TABLE):
@@ -1072,13 +1137,16 @@ def tag_fn(ctx, token):
     # Note: <nowiki> and HTML comments have already been handled in
     # preprocessing
 
-    # Do not try to parse HTML inside template arguments.  Many templates
-    # use tag-like constructions as internal markers, and they are not
-    # valid HTML tags.  HTML is only parsed after template processing.
-    if (_parser_have(ctx, NodeKind.TEMPLATE) or
-        _parser_have(ctx, NodeKind.TEMPLATE_ARG)):
+    # There are strings like <<country>> in some template arguments
+    if (token.startswith("<<") or
+        _parser_have(ctx, NodeKind.TEMPLATE) or
+        _parser_have(ctx, NodeKind.TEMPLATE_ARG) or
+        _parser_have(ctx, NodeKind.PARSER_FN)):
         return text_fn(ctx, token)
 
+    # If we are at the beginning of a line, close pending list
+    close_begline_lists(ctx)
+
     # Try to parse it as a start tag
     m = re.match(r"""<\s*([-a-zA-Z0-9]+)\s*((\b[-a-zA-Z0-9]+(=("[^"]*"|"""
                  r"""'[^']*'|[^ \t\n"'`=<>/]*))?\s*)*)(/?)\s*>""", token)
@@ -1089,6 +1157,14 @@ def tag_fn(ctx, token):
         also_end = m.group(6) == "/"
         name = name.lower()
 
+        # Some templates have markers like <1> in their arguments.  Only parse
+        # valid HTML tags in template arguments (tags like <math> can and
+        # do occur in them).
+        if (name not in ALLOWED_HTML_TAGS and
+            _parser_have(ctx, NodeKind.TEMPLATE) or
+            _parser_have(ctx, NodeKind.TEMPLATE_ARG)):
+            return text_fn(ctx, token)
+
         # If preparsing, only handle template control tags like <noinclude>
         if ctx.pre_parse:
             return text_fn(ctx, token)
@@ -1103,7 +1179,7 @@ def tag_fn(ctx, token):
             if also_end:
                 text_fn(ctx, MAGIC_NOWIKI_CHAR)
                 return
-            ctx.warning("unmatched <nowiki>")
+            ctx.debug("unmatched <nowiki>")
             return text_fn(ctx, token)
 
         # Handle <pre> start tag
@@ -1120,8 +1196,8 @@ def tag_fn(ctx, token):
         # tags that are allowed.
         if name not in ALLOWED_HTML_TAGS:
             if not name.isdigit() and not SILENT_HTML_LIKE:
-                ctx.warning("html tag <{}{}> not allowed in WikiText"
-                            "".format(name, "/" if also_end else ""))
+                ctx.debug("html tag <{}{}> not allowed in WikiText"
+                          "".format(name, "/" if also_end else ""))
             text_fn(ctx, token)
             return
 
@@ -1159,13 +1235,15 @@ def tag_fn(ctx, token):
 
     # Since it was not a start tag, it should be an end tag
     m = re.match(r"<\s*/\s*([-a-zA-Z0-9]+)\s*>", token)
-    assert m  # If fails, then mismatch between regexp here and tokenization
+    if m is None:
+        print("Could not match end tag token: {!r}".format(token))
+        assert False
     name = m.group(1)
     name = name.lower()
 
     # We should never see </section>
     if name == "section":
-        ctx.warning("unexpected </section>")
+        ctx.debug("unexpected </section>")
         return
 
     # Check for </pre> end tag
@@ -1174,7 +1252,7 @@ def tag_fn(ctx, token):
         ctx.pre_parse = False
         node = ctx.parser_stack[-1]
         if node.kind != NodeKind.PRE:
-            ctx.warning("unexpected </pre>")
+            ctx.debug("unexpected </pre>")
             return text_fn(ctx, token)
         _parser_pop(ctx, False)
         return
@@ -1186,8 +1264,8 @@ def tag_fn(ctx, token):
     # Give a warning on unsupported HTML tags.  WikiText limits the set of
     # tags that are allowed.
     if name not in ALLOWED_HTML_TAGS and name != "nowiki":
-        ctx.warning("html tag </{}> not allowed in WikiText"
-                    "".format(name))
+        ctx.debug("html tag </{}> not allowed in WikiText"
+                  "".format(name))
 
     # See if we can find the opening tag from the stack
     for i in range(0, len(ctx.parser_stack)):
@@ -1202,7 +1280,7 @@ def tag_fn(ctx, token):
             node.args = name
             _parser_pop(ctx, False)
             return
-        ctx.warning("no corresponding start tag found for {}".format(token))
+        ctx.debug("no corresponding start tag found for {}".format(token))
         text_fn(ctx, token)
         return
 
@@ -1232,6 +1310,7 @@ def tag_fn(ctx, token):
 
 def magicword_fn(ctx, token):
     """Handles a magic word, such as "__NOTOC__"."""
+    close_begline_lists(ctx)
     node = _parser_push(ctx, NodeKind.MAGIC_WORD)
     node.args = token
     _parser_pop(ctx, False)
@@ -1243,7 +1322,6 @@ token_re = re.compile(r"(?m)^(={2,6})\s*(([^=]|=[^=])+?)\s*(={2,6})\s*$|"
                       r"'''''|"
                       r"'''|"
                       r"''|"
-                      r"[ \t]+\n*|"
                       r"\n|"
                       r"\[|"
                       r"\]|"
@@ -1252,11 +1330,12 @@ token_re = re.compile(r"(?m)^(={2,6})\s*(([^=]|=[^=])+?)\s*(={2,6})\s*$|"
                       r"\|\+|"
                       r"\|-|"
                       r"!!|"
-                      r"^!|"
+                      r"^[ \t]*!|"
                       r"\|\||"
                       r"\||"
                       r"^----+|"
                       r"^[*:;#]+|"
+                      r"[ \t]+\n*|"
                       r":|"   # sometimes special when not beginning of line
                       r"<<[-a-zA-Z0-9/]*>>|"
                       r"""<\s*[-a-zA-Z0-9]+\s*(\b[-a-zA-Z0-9]+(=("[^<>"]*"|"""
@@ -1420,6 +1499,7 @@ def process_text(ctx, text):
     """Tokenizes ``text`` and processes each token in sequence.  This can be
     called recursively (which we do to process tokens inside templates and
     certain other structures)."""
+    # print("PARSER PROCESS_TEXT:", repr(text))
     for is_token, token in token_iter(ctx, text):
         # print("process_text: token_iter yielded:", is_token, token)
         node = ctx.parser_stack[-1]
@@ -1444,7 +1524,7 @@ def process_text(ctx, text):
                 subtitle_start_fn(ctx, token)
             elif token.startswith(">=="):  # Note: > added by tokenizer
                 subtitle_end_fn(ctx, token)
-            elif token.startswith("<"):
+            elif token.startswith("<"):  # HTML tag like construct
                 tag_fn(ctx, token)
             elif token.startswith("----"):
                 hline_fn(ctx, token)
@@ -1456,8 +1536,13 @@ def process_text(ctx, text):
                   ord(token) <= MAGIC_LAST):
                 magic_fn(ctx, token)
             else:
-                text_fn(ctx, token)
+                t2 = token.strip()
+                if t2 in tokenops:
+                    tokenops[t2](ctx, t2)
+                else:
+                    text_fn(ctx, token)
         ctx.linenum += token.count("\n")
+        ctx.wsp_beginning_of_line = ctx.beginning_of_line and token.isspace()
         ctx.beginning_of_line = token[-1] == "\n"
 
 
@@ -1469,6 +1554,7 @@ def parse_encoded(ctx, text):
     node = WikiNode(NodeKind.ROOT, 0)
     node.args.append([ctx.title])
     ctx.beginning_of_line = True
+    ctx.wsp_beginning_of_line = False
     ctx.linenum = 1
     ctx.pre_parse = False
     ctx.parser_stack = [node]
