@@ -1,6 +1,6 @@
 # Definitions for various parser functions supported in WikiText
 #
-# Copyright (c) 2020 Tatu Ylonen.  See file LICENSE and https://ylonen.org
+# Copyright (c) 2020, 2021 Tatu Ylonen.  See file LICENSE and https://ylonen.org
 
 import re
 import html
@@ -55,8 +55,8 @@ def if_fn(ctx, fn_name, args, expander):
     arg2 = args[2] if len(args) >= 3 else ""
     v = expander(arg0).strip()
     if v:
-        return expander(arg1).strip()
-    return expander(arg2).strip()
+        return expander(arg1)
+    return expander(arg2)
 
 
 def ifeq_fn(ctx, fn_name, args, expander):
@@ -66,8 +66,8 @@ def ifeq_fn(ctx, fn_name, args, expander):
     arg2 = args[2] if len(args) >= 3 else ""
     arg3 = args[3] if len(args) >= 4 else ""
     if expander(arg0).strip() == expander(arg1).strip():
-        return expander(arg2).strip()
-    return expander(arg3).strip()
+        return expander(arg2)
+    return expander(arg3)
 
 
 def iferror_fn(ctx, fn_name, args, expander):
@@ -78,10 +78,10 @@ def iferror_fn(ctx, fn_name, args, expander):
     if re.search(r'<[^>]*?\sclass="error"', arg0):
         if arg1 is None:
             return ""
-        return expander(arg1).strip()
+        return expander(arg1)
     if arg2 is None:
         return arg0
-    return expander(arg2).strip()
+    return expander(arg2)
 
 
 def ifexpr_fn(ctx, fn_name, args, expander):
@@ -105,8 +105,8 @@ def ifexist_fn(ctx, fn_name, args, expander):
     arg2 = args[2] if len(args) >= 3 else ""
     exists = ctx.page_exists(expander(arg0).strip())
     if exists:
-        return expander(arg1).strip()
-    return expander(arg2).strip()
+        return expander(arg1)
+    return expander(arg2)
 
 def switch_fn(ctx, fn_name, args, expander):
     """Implements #switch parser function."""
@@ -115,23 +115,22 @@ def switch_fn(ctx, fn_name, args, expander):
     defval = None
     last = None
     for i in range(1, len(args)):
-        arg = expander(args[i]).strip()
+        arg = args[i]
         m = re.match(r"(?s)^([^=]*)=(.*)$", arg)
         if not m:
-            last = arg
-            if arg == val:
+            last = expander(arg).strip()
+            if last == val:
                 match_next = True
             continue
         k, v = m.groups()
-        k = k.strip()
-        v = v.strip()
+        k = expander(k).strip()
         if k == val or match_next:
-            return v
+            return expander(v).strip()
         if k == "#default":
             defval = v
         last = None
     if defval is not None:
-        return defval
+        return expander(defval).strip()
     return last or ""
 
 
